/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.binding.soap.listener.incoming.jetty;

import java.net.SocketException;
import java.net.URI;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.axis2.AxisFault;
import org.apache.axis2.Constants;
import org.apache.axis2.addressing.EndpointReference;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.context.MessageContext;
import org.apache.axis2.context.SessionContext;
import org.apache.axis2.description.Parameter;
import org.apache.axis2.description.TransportInDescription;
import org.apache.axis2.engine.ListenerManager;
import org.apache.axis2.transport.TransportListener;
import org.apache.axis2.transport.http.server.HttpUtils;
import org.mortbay.jetty.Connector;
import org.mortbay.jetty.Server;
import org.mortbay.jetty.handler.ContextHandlerCollection;
import org.mortbay.jetty.nio.SelectChannelConnector;
import org.mortbay.jetty.servlet.Context;
import org.mortbay.jetty.servlet.ServletHolder;
import org.mortbay.thread.BoundedThreadPool;
import org.ow2.petals.binding.soap.SoapComponentContext;
import org.ow2.petals.binding.soap.listener.incoming.SoapServerConfig;
import org.ow2.petals.binding.soap.listener.incoming.servlet.ListServicesServlet;
import org.ow2.petals.binding.soap.listener.incoming.servlet.SoapServlet;
import org.ow2.petals.binding.soap.listener.incoming.servlet.SubscriptionServiceServlet;
import org.ow2.petals.ws.notification.WsnIsolationLayer;
import org.ow2.petals.ws.notification.WsnManager;

/**
 * Create a Jetty server that will be the axis2 transportlistener. This Jetty
 * server is used to replace the SimpleHTTPServer provided by axis that have
 * quite bad performances.
 *
 * @author Christophe Hamerling - EBM WebSourcing
 *
 */
public class SoapServer implements TransportListener {

    /**
     * The logger
     */
    protected Logger logger;

    /**
     * The axis2 configuration context (built from axis2.xml config file)
     */
    protected ConfigurationContext configurationContext;

    /**
     * The component properties
     */
    protected SoapServerConfig config;

    /**
     * The Jetty server
     */
    protected Server server;

    /**
     * The jetty thread pool. Size is configured from component configuration.
     */
    protected BoundedThreadPool threadPool;

    /**
     * The host address
     */
    protected String hostAddress = null;
    
    public static final int HEADER_BUFFER_SIZE = 4096 * 4; 

    /**
     *
     */
    protected WsnIsolationLayer isolation;

    static {
        JettyLogger.init();
    }

    /**
     * Creates a new instance of {@link SoapServer}
     *
     * @param configContext
     * @param config
     * @param componentContext
     * @param logger
     * @throws AxisFault
     */
    public SoapServer(ConfigurationContext configContext,
            SoapServerConfig config, SoapComponentContext componentContext,
            Logger logger) throws AxisFault {
        this.config = config;
        this.configurationContext = configContext;
        this.logger = logger;

        // jetty threapool configuration
        this.threadPool = new BoundedThreadPool();
        this.threadPool.setName("BCSoapJettyThreadPool");
        this.threadPool.setMaxThreads(this.config.getJettyThreadMaxPoolSize());
        this.threadPool.setMinThreads(this.config.getJettyThreadMinPoolSize());

        // jetty http connector configuration
        SelectChannelConnector nioConnector = new SelectChannelConnector();
        nioConnector.setPort(this.config.getPort());
        nioConnector.setHeaderBufferSize(HEADER_BUFFER_SIZE);
        nioConnector.setStatsOn(false);
        nioConnector.setAcceptors(this.config.getJettyAcceptors());

        this.server = new Server();
        this.server.setConnectors(new Connector[] { nioConnector });
        this.server.setThreadPool(threadPool);

        // create context handlers
        ContextHandlerCollection contexts = new ContextHandlerCollection();
        this.server.setHandler(contexts);

        // create the axis context
        Context axisContext = new Context(contexts, "/"
                + this.config.getServicesContext(), Context.SESSIONS);

        // create servlet holders
        ServletHolder axisServlet = new ServletHolder(new SoapServlet(
                configurationContext));
        axisServlet.setName("AxisServlet");
        axisServlet.setInitOrder(1);
        axisContext.addServlet(axisServlet, "/"
                + this.config.getServicesMapping() + "/*");
        axisContext.addServlet(axisServlet, "/servlet/AxisServlet");

        // add our own simple service list servlet
        ServletHolder listServlet = new ServletHolder(new ListServicesServlet(
                configurationContext, this.config.isProvidesList()));
        listServlet.setName("ServicesListServlet");
        listServlet.setInitOrder(1);
        axisContext.addServlet(listServlet, "/"
                + this.config.getServicesMapping() + "/listServices");

        // Notification servlet
        Context wsnContext = new Context(contexts, "/"
                + this.config.getWsnContext(), Context.SESSIONS);
        WsnManager manager = componentContext.getWsnManager();
        if (manager.getProducerEPR() == null) {
            manager
                    .setProducerEPR(new org.ow2.petals.ws.addressing.EndpointReference(
                            URI.create(this.config.buildWsnProducerURL())));
        }
        this.isolation = new WsnIsolationLayer(manager);
        ServletHolder subscribeServlet = new ServletHolder(
                new SubscriptionServiceServlet(isolation, logger));
        subscribeServlet.setName("subscribeServlet");
        wsnContext.addServlet(subscribeServlet, "/"
                + this.config.getWsnProducerMapping());

        configureListenerManager();
    }

    /**
     *
     * @return
     */
    public ConfigurationContext getConfigurationContext() {
        return this.configurationContext;
    }

    /*
     * (non-Javadoc)
     *
     * @see org.apache.axis2.transport.TransportListener#getEPRsForService(java.lang.String,
     *      java.lang.String)
     */
    public EndpointReference getEPRForService(String serviceName, String ip)
            throws AxisFault {
        return getEPRsForService(serviceName, ip)[0];
    }

    /*
     * (non-Javadoc)
     *
     * @see org.apache.axis2.transport.TransportListener#getEPRsForService(java.lang.String,
     *      java.lang.String)
     */
    public EndpointReference[] getEPRsForService(String serviceName, String ip)
            throws AxisFault {
        // if host address is present
        if (hostAddress != null) {
            if (this.server != null) {
                return new EndpointReference[] { new EndpointReference(
                        hostAddress + "/"
                                + configurationContext.getServiceContextPath()
                                + "/" + serviceName) };
            } else {
                throw new AxisFault(
                        "Unable to generate EPR for the transport : http");
            }
        }
        // if the host address is not present
        String localAddress;
        if (ip != null) {
            localAddress = ip;
        } else {
            try {
                localAddress = HttpUtils.getIpAddress();
                if (localAddress == null) {
                    localAddress = "127.0.0.1";
                }
            } catch (SocketException e) {
                throw AxisFault.makeFault(e);
            }
        }
        if (this.server != null) {
            return new EndpointReference[] { new EndpointReference("http://"
                    + localAddress + ":" + config.getPort() + "/"
                    + configurationContext.getServiceContextPath() + "/"
                    + serviceName) };
        } else {
            throw new AxisFault(
                    "Unable to generate EPR for the transport : http");
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.apache.axis2.transport.TransportListener#init(org.apache.axis2.context.ConfigurationContext,
     *      org.apache.axis2.description.TransportInDescription)
     */
    public void init(ConfigurationContext axisConf,
            TransportInDescription transprtIn) throws AxisFault {
        try {
            this.configurationContext = axisConf;

            Parameter param = transprtIn.getParameter(PARAM_PORT);

            if (param != null) {
                this.config
                        .setPort(Integer.parseInt((String) param.getValue()));
            }
            param = transprtIn.getParameter(HOST_ADDRESS);
            if (param != null) {
                hostAddress = ((String) param.getValue()).trim();
            }
        } catch (Exception e) {
            throw AxisFault.makeFault(e);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.apache.axis2.transport.TransportListener#start()
     */
    public void start() throws AxisFault {
        logger.info("Starting Jetty server...");
        logger.info("Port : " + config.getPort() + " / Max poolsize : "
                + config.getJettyThreadMaxPoolSize() + " / Min poolsize : "
                + config.getJettyThreadMinPoolSize() + "/ Acceptors size : "
                + config.getJettyAcceptors());
        try {
            this.server.start();
        } catch (Exception e) {
            logger.log(Level.SEVERE, "Can not start the Jetty server");
            throw new AxisFault("Can not start the Jetty server", e);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.apache.axis2.transport.TransportListener#stop()
     */
    public void stop() throws AxisFault {
        logger.log(Level.INFO, "Stopping Jetty server...");

        try {
            // stop jetty server
            this.server.stop();
        } catch (InterruptedException e) {
            throw AxisFault.makeFault(e);
        } catch (Exception e) {
            throw AxisFault.makeFault(e);
        }
    }

    /**
     * Configure the listener manager used by Axis
     *
     * @throws AxisFault
     */
    private void configureListenerManager() throws AxisFault {
        ListenerManager listenerManager = configurationContext
                .getListenerManager();
        TransportInDescription trsIn = new TransportInDescription(
                Constants.TRANSPORT_HTTP);
        trsIn.setReceiver(this);
        if (listenerManager == null) {
            listenerManager = new ListenerManager();
            listenerManager.init(configurationContext);
        }
        listenerManager.addListener(trsIn, true);
    }

    public SessionContext getSessionContext(MessageContext messageContext) {
        // TODO Auto-generated method stub
        return null;
    }

    public void destroy() {
        // TODO Auto-generated method stub
    }
}
