/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.binding.soap.util;

import static org.ow2.petals.binding.soap.Constants.JbiMessage.SOAP_HEADER;
import static org.ow2.petals.binding.soap.Constants.JbiMessage.SOAP_VERSION;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.Serializable;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.activation.DataHandler;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;

import org.apache.axiom.attachments.Attachments;
import org.apache.axiom.om.OMAttribute;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMNamespace;
import org.apache.axiom.om.OMNode;
import org.apache.axiom.om.OMText;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;
import org.apache.axiom.soap.SOAPBody;
import org.apache.axiom.soap.SOAPEnvelope;
import org.apache.axiom.soap.SOAPFactory;
import org.apache.axiom.soap.SOAPHeader;
import org.apache.axis2.AxisFault;
import org.apache.axis2.client.Options;
import org.apache.axis2.context.MessageContext;
import org.ow2.petals.binding.soap.listener.incoming.PetalsReceiver;
import org.ow2.petals.component.framework.util.Exchange;
import org.ow2.petals.component.framework.util.SourceHelper;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * A marshaller to create JBI message from SOAP ones and vice versa.
 *
 * @author Christophe HAMERLING (chamerling) - eBM WebSourcing
 *
 */
public class Marshaller {

    private static Marshaller INSTANCE;

    /**
     * Get the marshaller instance.
     *
     * @return
     */
    public synchronized static final Marshaller getInstance() {
        if (INSTANCE == null) {
            INSTANCE = new Marshaller();
        }
        return INSTANCE;
    }

    /**
     *
     * Creates a new instance of {@link Marshaller}
     *
     */
    private Marshaller() {
    }

    /**
     * Create a the JBI payload from the SOAP body. The SOAP envelope is
     * required to get namespaces.
     *
     * @param envelope
     * @param removeRootBodyElement
     * @return
     * @throws Exception
     */
    @SuppressWarnings("unchecked")
    public Source createSourceFromSOAPEnvelope(final SOAPEnvelope envelope,
            boolean removeRootBodyElement) throws Exception {
        Source result = null;
        final SOAPBody body = envelope.getBody();
        final OMNamespace namespace = envelope.getNamespace();
        final Iterator<OMNamespace> envNS = envelope.getAllDeclaredNamespaces();
        final Iterator<OMNamespace> bodyNS = body.getAllDeclaredNamespaces();

        OMElement rootElement = body.getFirstElement();
        if (removeRootBodyElement) {
            rootElement = rootElement.getFirstElement();
        }

        if (rootElement != null) {
            rootElement.declareNamespace(namespace);
            while (envNS.hasNext()) {
                rootElement.declareNamespace(envNS.next());
            }
            while (bodyNS.hasNext()) {
                rootElement.declareNamespace(bodyNS.next());
            }

            ByteArrayOutputStream os = new ByteArrayOutputStream();
            rootElement.serialize(os);

            os.toByteArray();
            result = new StreamSource(
                    new ByteArrayInputStream(os.toByteArray()));
        }
        return result;
    }

    /**
     * Put the SOAP attachments in the normalized message
     *
     * @param attachments
     * @param message
     */
    @SuppressWarnings("unchecked")
    public void copySOAPAttachmentsToNM(Attachments attachments,
            NormalizedMessage message) {
        if (message != null && attachments != null) {
            if (attachments != null) {
                Set set = attachments.getContentIDSet();
                for (Object object : set) {
                    String id = (String) object;
                    DataHandler dh = attachments.getDataHandler(id);
                    try {
                        message.addAttachment(id, dh);
                    } catch (MessagingException e) {
                    }
                }
            }
        }
    }

    /**
     * Copy properties from a SOAP message to the message exchange
     *
     * @param context
     * @param exchange
     */
    @SuppressWarnings("unchecked")
    public void copySOAPPropertiesToME(MessageContext context,
            Exchange exchange) {
        Options options = context.getOptions();

        // get serializable properties, and add them to the message exchange
        // properties
        Map properties = options.getProperties();
        for (Object key : properties.keySet()) {
            Object value = properties.get(key);
            if (value != null) {
                // for a collection, we must check if all objects are serializable
                if (value instanceof Collection) {
                    if (!((Collection)value).isEmpty()) {
                        boolean allSerializables = true;
                        for (Object elementValue : ((Collection)value).toArray()) {
                            if (!(elementValue instanceof Serializable)) {
                                allSerializables = false;
                                break;
                            }
                        }
                        if (allSerializables) {
                            exchange.setProperty(key.toString(), properties.get(key));
                        }
                    }
                } else if (value instanceof Serializable) {
                    exchange.setProperty(key.toString(), properties.get(key));
                }
            }
        }

        // get the SOAP header from envelope and add it as exchange property if
        // it exists
        try {
            SOAPEnvelope env = context.getEnvelope();
            SOAPHeader header = env.getHeader();
            if (header != null) {
                ByteArrayOutputStream os = new ByteArrayOutputStream();
                header.serialize(os);
                exchange.setProperty(SOAP_HEADER, os.toString());
            }
        } catch (XMLStreamException e) {
        }

        // get additional options
        exchange.setProperty(SOAP_VERSION, options.getSoapVersionURI());
    }

    /**
     *
     * @param factory
     * @param envelope
     * @param response
     * @return
     * @throws AxisFault
     */
    public SOAPBody fillSOAPBodyFromSource(SOAPFactory factory,
            SOAPEnvelope envelope, Source jbiPayload) throws AxisFault {

        try {
        	XMLStreamReader parser = StaxUtils.createXMLStreamReader(jbiPayload);
            StAXOMBuilder builder = new StAXOMBuilder(factory, parser);
	        OMElement bodyContent = builder.getDocumentElement();

            SOAPBody respBody = factory.createSOAPBody(envelope);
        	respBody.addChild(bodyContent);
        	return respBody;

        } catch (Exception e) {
            throw new AxisFault(
                    "Error parsing the response from JBI service to a SOAPBody",
                    PetalsReceiver.FAULT_SERVER,
                    e);
        }
    }

    /**
     * Creates a SOAP response from a NormalizedMessage
     *
     * @param factory
     *            soap factory
     * @param nm
     *            NormalizedMessage containing the response
     * @return a SOAPEnveloppe created from the nm NomalizedMessage content
     * @throws AxisFault
     */
    public SOAPEnvelope createSOAPEnvelopeFromNM(SOAPFactory factory,
            NormalizedMessage nm) throws AxisFault {
        Source source = null;
        SOAPEnvelope responseEnv = factory.createSOAPEnvelope();

        /*
         * Create and fill the Soap body with the content of the Normalized
         * message
         */
        if (nm == null || nm.getContent() == null) {
            try {
                final DocumentBuilderFactory documentFactory = DocumentBuilderFactory.newInstance();
                documentFactory.setNamespaceAware(true);
                
                final DocumentBuilder documentBuilder = documentFactory.newDocumentBuilder();
                Document document = documentBuilder.newDocument();
                
                final Element responseElement = document.createElement("Response");
                responseElement.setNodeValue("Done");
                document.appendChild(responseElement);
                source = SourceHelper.createSource(document);
            } catch (ParserConfigurationException e) {
                throw new AxisFault(e.getMessage(), "Server", e);
            }
        } else {
            source = nm.getContent();
        }

        this.fillSOAPBodyFromSource(factory, responseEnv, source);
        return responseEnv;
    }

    /**
     * Add the normalized message attachments to the Axis message context.
     *
     * @param nm
     * @param messageContext
     *
     * @throws AxisFault
     */
    @SuppressWarnings("unchecked")
    public void copyJBIAttachmentsToSOAP(NormalizedMessage nm,
            MessageContext messageContext) {
        if (nm != null && messageContext != null) {
            Set attachmentNames = nm.getAttachmentNames();
            for (Object id : attachmentNames) {
                String name = (String) id;
                messageContext.addAttachment(name, nm.getAttachment(name));
            }
        }
    }

    /**
     * Create the JBI message from the OMElement object.
     *
     * @param omElement
     * @param message
     */
    @SuppressWarnings("unchecked")
    public void copySOAPAttachementTOJBI(OMElement omElement,
            NormalizedMessage message) {

        // get attachments
        Iterator iter = omElement.getChildren();
        while (iter.hasNext()) {
            OMNode node = (OMNode) iter.next();
            if (node instanceof OMElement) {
                OMElement element = (OMElement) node;
                // (all the nodes that have an href attributes are
                // attachments)
                OMAttribute attr = element.getAttribute(new QName("href"));
                if ((attr != null) && (node instanceof OMText)) {
                    OMText binaryNode = (OMText) node;
                    DataHandler dh = (DataHandler) binaryNode.getDataHandler();
                    try {
                        message.addAttachment(dh.getName(), dh);
                    } catch (MessagingException e) {
                    }
                }
            }
        }
    }
}
